SystemOrganization addCategory: #'AmbientTalkST-Core'!

!Behavior methodsFor: '*AmbientTalkST-Core-accessing' stamp: 'StefanMarr 1/2/2012 16:18'!
methodDict
	methodDict == nil ifTrue: [^ self emptyMethodDictionary].
	^ methodDict! !

Process subclass: #ATActor
	instanceVariableNames: 'mailbox terminate'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AmbientTalkST-Core'!

!ATActor classSide methodsFor: 'instance creation' stamp: 'StefanMarr 1/23/2013 11:12'!
new
	"Create a new actor and start it."
	| newActor |
	newActor := super new.
	newActor suspendedContext: [
		newActor start.
		newActor terminateProcess.
	] asContext.
	newActor priority: Processor activePriority.
	newActor resume.
	^ newActor! !

!ATActor classSide methodsFor: 'instance creation' stamp: 'StefanMarr 6/1/2012 16:19'!
newWithInitializer: aBlock
	"Normally, we would return a reference to the domain,
	 but this is not really the way AmbientTalk works,
	 so, here we also return the value returned by the aBlock."
	| domain valueMsg promise |
	domain := self new.
	promise := domain evaluateAsync: aBlock.
	^ domain adoptObject: promise read.! !

!ATActor methodsFor: 'actor semantics' stamp: 'StefanMarr 1/3/2012 15:09'!
adoptObject: anObj
	^ ATFarReference for: anObj in: self.! !

!ATActor methodsFor: 'actor semantics' stamp: 'StefanMarr 1/1/2012 17:08'!
enqueue: aMessage
	"Directly put a message into the actors mailbox."
	mailbox nextPut: aMessage! !

!ATActor methodsFor: 'actor semantics' stamp: 'StefanMarr 12/31/2011 15:09'!
evaluateAsync: aBlock
	"This is an implementation level method.
	 The block is by convention not referring to
	 unsafe shared state."
	| msg |
	msg := ATMessage new.
	msg receiver: aBlock.
	msg setSelector: #value arguments: #().

	mailbox nextPut: msg.
	
	^ msg promise.! !

!ATActor methodsFor: 'initialization' stamp: 'StefanMarr 1/2/2012 11:51'!
initialize
	super initialize.
	mailbox := SharedQueue new.
	
	terminate := false.! !

!ATActor methodsFor: 'printing' stamp: 'StefanMarr 1/4/2012 14:32'!
printOn: aStream
	name 	ifNil:    [ aStream nextPutAll: 'Actor']
			ifNotNil: [ aStream nextPutAll: 'Actor [', name, ']'.].
	super printOn: aStream.! !

!ATActor methodsFor: 'actor semantics' stamp: 'StefanMarr 5/31/2013 14:39'!
processIncomingMessages
	| msg result |
	msg := mailbox next.
	
	[ result := msg send.
	  msg result: result ] on: Error do: [:err | msg failed: err. Halt now. ]! !

!ATActor methodsFor: 'private' stamp: 'StefanMarr 1/2/2012 11:50'!
start
	[terminate] whileFalse:
		[ self processIncomingMessages. ]! !

!ATActor methodsFor: 'actor semantics' stamp: 'StefanMarr 1/23/2013 13:04'!
terminate
	"Return promise to be able to await termination"
	^ self evaluateAsync: [ terminate := true ]! !

!ATActor methodsFor: 'private' stamp: 'StefanMarr 1/23/2013 11:14'!
terminateProcess
	"Helper method to access the super-class' terminate."
	super terminate! !

Object subclass: #ATFarReferenceBase
	instanceVariableNames: 'obj actor'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AmbientTalkST-Core'!

!ATFarReferenceBase commentStamp: 'StefanMarr 1/1/2012 16:45' prior: 0!
ATFarReferenceBase is providing the data and #cannotInterpret: for ATFarReference.
Since #cannotInterpret: is always send to the super class, we needed this base class.
Refer to ATFarReference for relevant class-side methods.

Instance Variables:
	obj	   - the object this far-reference is pointing to
	actor  - the actor the referenced object is contained in!

ATFarReferenceBase subclass: #ATFarReference
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AmbientTalkST-Core'!

!ATFarReference commentStamp: 'StefanMarr 1/1/2012 16:51' prior: 0!
ATFarReference implements a far-reference, i.e., a proxy for an object that is referenced from another actor.
The base class ATFarReferenceBase provides the proxy implementation.
The methodDict of ATFarReference is set to nil to be able to use the #cannotInterpet: mechanism, which enables us
to intercept all messages sent to an object, including #doesNotUnderstand.

See the class-side methods to instantiate a far-reference.!

!ATFarReference classSide methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/1/2012 21:45'!
deStratify
	self methodDict: self emptyMethodDictionary.! !

!ATFarReference classSide methodsFor: 'as yet unclassified' stamp: 'StefanMarr 8/27/2012 16:03'!
for: anObj in: anActor
	| ref |
	"SmallIntegers and bools are immutable and directly used."
	(((((anObj class == ATFarReference)
	  or: [anObj isKindOf: SmallInteger])
	  or: [anObj isKindOf: Boolean])
	  or: [anObj isKindOf: Character])
	  or: [anObj isNil])
		ifTrue: [^anObj].
	
	"Strings and Floats are copied as by-value objects kind of isolates"
	((anObj isKindOf: String)
	  or: [anObj isKindOf: Float])
		ifTrue: [^ anObj copy].
		
	ref := ATFarReferenceBase new.
	ref object: anObj.
	ref actor:  anActor.
	
	ATFarReference adoptInstance: ref.
	
	^ ref! !

!ATFarReference classSide methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/2/2012 14:03'!
initialize
	"self initialize."

	super initialize.
	self stratify.! !

!ATFarReference classSide methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/1/2012 22:06'!
isStratified
	^ methodDict isNil! !

!ATFarReference classSide methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/2/2012 13:07'!
stratify
	self methodDict: nil.! !

!ATFarReferenceBase methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/1/2012 16:22'!
actor: anActor
	actor := anActor! !

!ATFarReferenceBase methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/22/2013 17:55'!
cannotInterpret: aMessage
	| msg |
	[Transcript cr; show: '#ci: rcvr: ', obj asString, ' msg: ', aMessage asString].

	"In case we have a far reference to a local object, we do it directly.
	 THIS SHOULD NEVER HAPPEN!!!!!!"
	actor == Processor activeProcess ifTrue: [
		"Transcript cr; show: 'THIS SHOULD NEVER HAPPEN msg:', aMessage asString, ' on: ', obj asString, ' in: ', actor asString."	
		^ aMessage sendTo: obj.
	].
	
	aMessage selector = #unwrapATRef   ifTrue: [^ obj   ].
	aMessage selector = #unwrapATActor ifTrue: [^ actor ].
	aMessage selector = #isATFarRef    ifTrue: [^ true  ].
	
	msg := ATMessage from: aMessage.
	msg receiver: obj.
	msg unwrapFarReferencesFor: actor.
	actor enqueue: msg.
	
	^ ATFarReference for: msg result in: actor.! !

!ATFarReferenceBase methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/2/2012 13:07'!
doesNotUnderstand: aMessage 
	"Try this for debugging, and for use in non-stratified mode."
	^ self cannotInterpret: aMessage! !

!ATFarReferenceBase methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/22/2013 17:55'!
isATFarRef
	^ true! !

!ATFarReferenceBase methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/1/2012 16:22'!
object: anObj
	obj := anObj! !

!ATFarReferenceBase methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/2/2012 13:59'!
unwrapATActor
	^ actor! !

!ATFarReferenceBase methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/2/2012 13:46'!
unwrapATRef
	^ obj! !

Object subclass: #ATPromise
	instanceVariableNames: 'signal value terminatedWithError'
	classVariableNames: 'NoValueDelivered'
	poolDictionaries: ''
	category: 'AmbientTalkST-Core'!

!ATPromise classSide methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/31/2011 14:20'!
example
	"This is not a real test case, but just to see how it is supposed to work"
	| p |
	p := ATPromise new.
	
	[|val| 
	 Transcript log: 'About to read promised value.'.
	 val := p read.
	 Transcript log: 'Val: ', val asString.] fork.

	Processor yield.
	
	Transcript log: 'About to deliver promised value.'.
	p deliver: 42.
	Transcript log: 'Deliver promised value.'.
	! !

!ATPromise classSide methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/4/2011 15:53'!
initialize
	NoValueDelivered := Object new.! !

!ATPromise methodsFor: 'accessing' stamp: 'StefanMarr 5/30/2013 16:36'!
await
	| i |
	i := 0.
	
	[value = NoValueDelivered and: [terminatedWithError = nil]] whileTrue: [
		"This is ugly, but I do not see how I can prevent the data
		 race for signals otherwise.
		 In case there are multiple possible consumers everyone
		 but the first will starve when they arrive before the
		 promise was delivered.
		 Now, we just have a delay, but it is guaranteed to return."
		signal waitTimeoutMSecs: 100.
		i := i + 1.
		i > 300 ifTrue: [Halt now].
	].! !

!ATPromise methodsFor: 'announcing' stamp: 'StefanMarr 11/4/2011 16:07'!
deliver: aValue
	"Deliver the promised value."
	
	
	"Warning: this is racy, but we don't care for the moment.
	 Should be a single produce anyway. Everything else is a bug.
	 And, we do at least best effort error notification..."
	value = NoValueDelivered ifFalse: [
		| err |
		err := Error new.
		err messageText: 'Promise has been delivered more than once.'.
		err signal.
	].

	value := aValue.
	
	signal signal.
	
	^ value.! !

!ATPromise methodsFor: 'announcing' stamp: 'StefanMarr 1/22/2013 17:12'!
failed: anError
	terminatedWithError := anError! !

!ATPromise methodsFor: 'initialization' stamp: 'StefanMarr 11/4/2011 16:09'!
initialize
	value := NoValueDelivered.
	signal := Semaphore new.! !

!ATPromise methodsFor: 'accessing' stamp: 'StefanMarr 1/22/2013 17:14'!
read
	self await.
	terminatedWithError = nil ifTrue: [
		^ value ].
	
	"On failure, raise the exception again in the current context."
	terminatedWithError signal! !

Message subclass: #ATMessage
	instanceVariableNames: 'sender receiver result'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AmbientTalkST-Core'!

!ATMessage classSide methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/1/2012 17:06'!
from: aMessage
	"Used to convert a normal Message object to an ATMessage object"
	| newATMsg |
	newATMsg := self new.
	newATMsg setSelector: aMessage selector arguments: aMessage arguments.
	newATMsg lookupClass: aMessage lookupClass.
	^ newATMsg.! !

!ATMessage methodsFor: 'accessing' stamp: 'StefanMarr 1/22/2013 17:11'!
failed: anError
	result failed: anError! !

!ATMessage methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/31/2011 15:11'!
initialize
	super initialize.
	result := ATPromise new.! !

!ATMessage methodsFor: 'accessing' stamp: 'StefanMarr 11/25/2011 17:20'!
promise
	^ result! !

!ATMessage methodsFor: 'accessing' stamp: 'StefanMarr 11/25/2011 17:15'!
receiver
	^ receiver! !

!ATMessage methodsFor: 'accessing' stamp: 'StefanMarr 11/25/2011 17:15'!
receiver: anObj
	receiver := anObj! !

!ATMessage methodsFor: 'accessing' stamp: 'StefanMarr 1/1/2012 21:48'!
result
	"Result is blocking on the promise.
	 Use #promise to obtain the promise directly."
	^ result read! !

!ATMessage methodsFor: 'accessing' stamp: 'StefanMarr 11/25/2011 17:20'!
result: anObj
	result deliver: anObj! !

!ATMessage methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/1/2012 20:38'!
send
	"answer the result of sending this message to its receiver"

	^ receiver perform: selector withArguments: args! !

!ATMessage methodsFor: 'accessing' stamp: 'StefanMarr 11/25/2011 17:15'!
sender
	^ sender! !

!ATMessage methodsFor: 'accessing' stamp: 'StefanMarr 11/25/2011 17:15'!
sender: anObj
	sender := anObj! !

!ATMessage methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/2/2012 14:00'!
unwrapFarReferencesFor: anActor
	args withIndexDo: [:val :i |
		((val class == ATFarReference) and: [val unwrapATActor == anActor])
			ifTrue: [args at: i put: val unwrapATRef].
	].! !
ATFarReference initialize!
ATPromise initialize!
